@@ -2,7 +2,7 @@
 #
 # The MIT License
 # 
-# Copyright (c) 2008 Felix Schwarz <felix.schwarz@oss.schwarz.eu>
+# Copyright (c) 2008-2009 Felix Schwarz <felix.schwarz@oss.schwarz.eu>
 # 
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -25,7 +25,10 @@
 import base64
 import binascii
 import re
-from sets import Set
+try:
+    set
+except NameError:
+    from sets import Set as set
 
 from repoze.workflow.statemachine import StateMachine, StateMachineError
 
@@ -48,7 +51,7 @@ class PyMTAException(Exception):
         self.reply_text = reply_text
 
 class DynamicStateSwitchException(PyMTAException):
-    """Used to implement more complicated state transistions where the next 
+    """Used to implement more complicated state transitions where the next 
     state depends on other values (e.g. where to go after a message was 
     received - depends on ESMTP and AUTH)."""
     def __init__(self, new_state):
@@ -65,15 +68,18 @@ class PolicyDenial(PyMTAException):
 
 class SMTPSession(object):
     """The SMTPSession processes all input data which were extracted from 
-    sockets previously. The idea behind is that this class is decoupled from 
-    asynchat as much as possible and make it really testable.
+    sockets previously. The idea behind this is that this class only knows about
+    different SMTP commands and does not have to know things like command mode
+    and data mode.
     
     The protocol parser will create a new session instance for every new 
     connection so this class does not have to be thread-safe.
     """
     
-    def __init__(self, command_parser, policy=None, authenticator=None):
+    def __init__(self, command_parser, deliverer, policy=None, 
+                 authenticator=None):
         self._command_parser = command_parser
+        self._deliverer = deliverer
         self._policy = policy
         self._authenticator = authenticator
         
@@ -84,13 +90,14 @@ class SMTPSession(object):
         self._build_state_machine()
     
     # -------------------------------------------------------------------------
+    # State machine building
     
     def _add_state(self, from_state, smtp_command, to_state):
         handler_function = self._dispatch_commands
         self.state.add(from_state, smtp_command, to_state, handler_function)
     
     def _get_all_real_states(self, including_quit=False):
-        states = Set()
+        states = set()
         for key in self.state.states:
             command_name = key[1]
             new_state = self.state.states[key]
@@ -101,7 +108,7 @@ class SMTPSession(object):
         return states
     
     def get_all_allowed_internal_commands(self):
-        """Returns an interable which includes all allowed commands. This does
+        """Returns an iterable which includes all allowed commands. This does
         not mean that a specific command from the result is executable right now
         in this session state (or that it can be executed at all in this 
         connection).
@@ -109,14 +116,14 @@ class SMTPSession(object):
         Please note that the returned values are /internal/ commands, not SMTP
         commands (use get_all_allowed_smtp_commands for that) so there will be
         'MAIL FROM' instead of 'MAIL'."""
-        states = Set()
+        states = set()
         for command_name, invalid in self._get_all_real_states(including_quit=True):
             if command_name not in ['GREET', 'MSGDATA']:
                 states.add(command_name)
         return states
     
     def get_all_allowed_smtp_commands(self):
-        states = Set()
+        states = set()
         for command_name in self.get_all_allowed_internal_commands():
             command_name = command_name.split(' ')[0]
             states.add(command_name)
@@ -132,7 +139,7 @@ class SMTPSession(object):
     def _add_help_noop_and_quit_transitions(self):
         """HELP, NOOP and QUIT should be possible from everywhere so we 
         need to add these transitions to all states configured so far."""
-        states = Set()
+        states = set()
         for key in self.state.states:
             new_state = self.state.states[key]
             state_name = new_state[0]
@@ -155,7 +162,7 @@ class SMTPSession(object):
         self._add_state('initialized', 'MAIL FROM',  'sender_known')
         self._add_state('esmtp_initialized', 'MAIL FROM',  'sender_known')
         
-        self._add_state('esmtp_initialized', 'AUTH PLAIN',  'authenticated')
+        self._add_state('esmtp_initialized', 'AUTH PLAIN', 'authenticated')
         self._add_state('authenticated', 'MAIL FROM',  'sender_known')
         # ----
         
@@ -168,6 +175,25 @@ class SMTPSession(object):
         self._add_rset_transitions()
         self.valid_commands = [command for from_state, command in self.state.states]
     
+    # -------------------------------------------------------------------------
+    
+    def get_ehlo_lines(self):
+        """Return the capabilities to be advertised after EHLO."""
+        lines = []
+        if self._authenticator != None:
+            # TODO: Make the authentication pluggable but separate mechanism 
+            # from user look-up.
+            lines.append('AUTH PLAIN')
+        if self._policy != None:
+            lines.extend(self._policy.ehlo_lines(self._message.peer))
+        lines.append('HELP')
+        return lines
+    
+    def _set_size_restrictions(self):
+        """Set the maximum allowed message in the underlying layer so that big 
+        messages are not hold in memory before they are rejected."""
+        max_message_size = self._get_max_message_size_from_policy()
+        self._command_parser.set_maximum_message_size(max_message_size)
     
     def _dispatch_commands(self, from_state, to_state, smtp_command, ob):
         """This method dispatches a SMTP command to the appropriate handler 
@@ -221,6 +247,7 @@ class SMTPSession(object):
         if decision:
             if not response_sent:
                 self.handle_input('greet')
+            self._set_size_restrictions()
         else:
             if not response_sent:
                 self.reply(554, 'SMTP service not available')
@@ -243,23 +270,26 @@ class SMTPSession(object):
                 msg = 'Command "%s" is not allowed here' % smtp_command
                 allowed_transitions = self.state.transitions(self)
                 if len(allowed_transitions) > 0:
-                      msg += ', expected on of %s' % allowed_transitions
+                    msg += ', expected on of %s' % allowed_transitions
                 self.reply(503, msg)
         except InvalidParametersException, e:
             if not e.response_sent:
                 msg = 'Syntactically invalid %s argument(s)' % smtp_command
                 self.reply(501, msg)
         except PolicyDenial, e:
-            print 'checking for response sent', e.response_sent
             if not e.response_sent:
                 self.reply(e.code, e.reply_text)
         self._command_arguments = None
     
+    def input_exceeds_limits(self):
+        """Called when the client sent a message that exceeded the maximum 
+        size."""
+        self.reply(552, 'message exceeds fixed maximum message size')
     
     def reply(self, code, text):
         """This method returns a message to the client (actually the session 
         object is responsible of actually pushing the bits)."""
-        print 'code, text', code, text
+        #print 'code, text', code, text
         self._command_parser.push(code, text)
     
     
@@ -324,7 +354,8 @@ class SMTPSession(object):
         self._message.smtp_helo = helo_string
         if not response_sent:
             primary_hostname = self._command_parser.primary_hostname
-            self.multiline_reply(250, (primary_hostname, 'AUTH PLAIN'))
+            lines = [primary_hostname] + self.get_ehlo_lines()
+            self.multiline_reply(250, lines)
     
     def smtp_ehlo(self):
         self._process_helo_or_ehlo('accept_ehlo', self._reply_to_ehlo)
@@ -359,10 +390,46 @@ class SMTPSession(object):
             else:
                 raise InvalidParametersException(credentials)
     
+    def _split_mail_from_parameter(self, data):
+        sender = data
+        extensions = {}
+        
+        # TODO: case insensitivity of extension names
+        verb_regex = re.compile('^\s*<(.*)>(?:\s*(\S+)\s*)*\s*$')
+        match = verb_regex.search(data)
+        if match:
+            sender = match.group(1)
+            extension_string = match.group(2)
+            if extension_string is not None:
+                for extension in re.split('\s+', extension_string):
+                    if '=' in extension:
+                        name, parameter = extension.split('=', 1)
+                        extensions[name] = parameter
+                    else:
+                        extensions[extension] = True
+        return (sender, extensions)
+    
+    def _check_mail_extensions(self, extensions):
+        if 'size' in extensions:
+            # TODO: protect against non-numeric size!
+            announced_size = int(extensions['size'])
+            max_message_size = self._get_max_message_size_from_policy()
+            if max_message_size != None:
+                if announced_size > max_message_size:
+                    self.reply(552, 'message exceeds fixed maximum message size')
+                    raise InvalidParametersException('MAIL FROM', response_sent=True)
+    
     def smtp_mail_from(self):
-        sender = self._command_arguments
+        data = self._command_arguments
+        sender, extensions = self._split_mail_from_parameter(data)
         # TODO: Check for good email address!
         # TODO: Check for single email address!
+        uses_esmtp = (self._state in ['esmtp_initialized', 'authenticated'])
+        if uses_esmtp:
+            self._check_mail_extensions(extensions)
+        elif len(extensions) > 0:
+            self.reply(501, 'No SMTP extensions allowed for plain SMTP')
+            raise InvalidParametersException('MAIL', response_sent=True)
         decision, response_sent = self.is_allowed('accept_from', sender, self._message)
         if decision:
             self._message.smtp_from = sender
@@ -398,6 +465,20 @@ class SMTPSession(object):
         elif not decision:
             raise PolicyDenial(response_sent)
     
+    def _get_max_message_size_from_policy(self):
+        max_message_size = None
+        if (self._policy is not None) and (self._message.peer is not None):
+            max_message_size = self._policy.max_message_size(self._message.peer)
+        return max_message_size
+    
+    def _check_size_restrictions(self, msg_data):
+        max_message_size = self._get_max_message_size_from_policy()
+        if (max_message_size is not None):
+            msg_too_big = (len(msg_data) > int(max_message_size))
+            if msg_too_big:
+                msg = 'message exceeds fixed maximum message size'
+                raise PolicyDenial(False, 552, msg)
+    
     def _copy_basic_settings(self, msg):
         peer = self._message.peer
         new_message = Message(peer=Peer(peer.remote_ip, peer.remote_port), 
@@ -410,11 +491,12 @@ class SMTPSession(object):
         whole message was received (multi-line DATA command is completed)."""
         msg_data = self._command_arguments
         self._command_parser.switch_to_command_mode()
+        self._check_size_restrictions(msg_data)
         decision, response_sent = self.is_allowed('accept_msgdata', msg_data, self._message)
         if decision:
             self._message.msg_data = msg_data
             new_message = self._copy_basic_settings(self._message)
-            self._command_parser.new_message_received(self._message)
+            self._deliverer.new_message_accepted(self._message)
             if not response_sent:
                 self.reply(250, 'OK')
                 # Now we must not loose the message anymore!
